import base64
from typing import List


def i2osp(x: int, x_len: int) -> bytes:
    """
    Integer-to-Octet-String primitive
    """
    if x >= 256 ** x_len:
        raise ValueError("integer too large")
    digits = []
    while x:
        digits.append(int(x % 256))
        x //= 256
    for i in range(x_len - len(digits)):
        digits.append(0)
    return bytes.fromhex("".join("%.2x" % x for x in digits[::-1]))


def os2ip(octet_string: bytes) -> int:
    """
    Octet-String-to-Integer primitive
    """
    x_len = len(octet_string)
    octet_string = octet_string[::-1]
    x = 0
    for i in range(x_len):
        x += octet_string[i] * 256 ** i
    return x


def base64url_decode(v: str) -> bytes:
    bv = v.encode("ascii")
    rem = len(bv) % 4
    if rem > 0:
        bv += b"=" * (4 - rem)
    return base64.urlsafe_b64decode(bv)


def base64url_encode(input: bytes) -> bytes:
    return base64.urlsafe_b64encode(input).replace(b"=", b"")


def _le64(n: int) -> bytes:
    s = bytearray(8)
    for i in range(8):
        if i == 7:
            n = n & 127
        s[i] = n & 255
        n = n >> 8
    return bytes(s)


def pae(pieces: List[bytes]) -> bytes:
    output = _le64(len(pieces))
    for i in range(len(pieces)):
        output += _le64(len(pieces[i]))
        output += pieces[i]
    return output



import json

import pytest

import pyseto
from pyseto import Key

from .utils import get_path


def _load_tests(paths: list) -> list:
    tests: list = []
    for path in paths:
        with open(get_path(path)) as tv_file:
            tv = json.loads(tv_file.read())
        tests += tv["tests"]
    return tests


class TestWithTestVectors:
    """
    Tests with test vectors defined in https://github.com/paseto-standard/test-vectors.
    """

    @pytest.mark.parametrize(
        "v",
        _load_tests(
            [
                "vectors/v1.json",
                "vectors/v2.json",
                "vectors/v3.json",
                "vectors/v4.json",
            ]
        ),
    )
    def test_with_test_vectors(self, v):

        token = v["token"].encode("utf-8")
        payload = json.dumps(v["payload"], separators=(",", ":")).encode("utf-8")
        footer = v["footer"].encode("utf-8")
        implicit_assertion = v["implicit-assertion"].encode("utf-8")

        version = v["name"].split("-")[0]
        purpose = v["name"].split("-")[1]
        if purpose == "E":
            nonce = bytes.fromhex(v["nonce"])
            key = bytes.fromhex(v["key"])

            k = Key.new("v" + version, "local", key=key)
            encoded = pyseto.encode(k, payload, footer, implicit_assertion, nonce=nonce)
            decoded_token = pyseto.decode(k, token, implicit_assertion)
            decoded = pyseto.decode(k, encoded, implicit_assertion)
            assert payload == decoded_token.payload == decoded.payload
            return

        if purpose == "S":
            secret_key_pem = v["secret-key"] if version == "1" else v["secret-key-pem"]
            public_key_pem = v["public-key"] if version == "1" else v["public-key-pem"]

            sk = Key.new("v" + version, "public", secret_key_pem)
            encoded = pyseto.encode(sk, payload, footer, implicit_assertion)
            pk = Key.new("v" + version, "public", public_key_pem)
            decoded_token = pyseto.decode(pk, token, implicit_assertion)
            decoded = pyseto.decode(pk, encoded, implicit_assertion)
            assert payload == decoded_token.payload == decoded.payload
            return
        pytest.fail(f"Invalid test name: {v['name']}")



from typing import Any, Union

from cryptography import x509
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)

from .exceptions import NotSupportedError


class KeyInterface:
    """
    The key interface class for PASETO.
    """

    def __init__(self, version: str, type: str, key: Union[bytes, str]):
        self._version = version
        self._type = type
        self._header = (self._version + "." + self._type + ".").encode("utf-8")
        self._sig_size = 0
        self._key: Any = None

        bkey = key if isinstance(key, bytes) else key.encode("utf-8")
        if self._type == "local":
            self._key = bkey
            return

        skey = key if isinstance(key, str) else key.decode("utf-8")
        if "BEGIN CERTIFICATE" in skey:
            self._key = x509.load_pem_x509_certificate(bkey).public_key()
        elif "BEGIN EC PRIVATE" in skey:
            self._key = load_pem_private_key(bkey, password=None)
        elif "BEGIN PRIVATE" in skey:
            self._key = load_pem_private_key(bkey, password=None)
        elif "BEGIN PUBLIC" in skey:
            self._key = load_pem_public_key(bkey)
        elif "BEGIN RSA PRIVATE" in skey:
            self._key = load_pem_private_key(bkey, password=None)
        else:
            raise ValueError("Failed to decode PEM.")
        return

    @property
    def version(self) -> str:
        return self._version

    @property
    def type(self) -> str:
        return self._type

    @property
    def header(self) -> bytes:
        return self._header

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:
        raise NotSupportedError()

    def decrypt(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        raise NotSupportedError()

    def sign(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        raise NotSupportedError()

    def verify(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        raise NotSupportedError()



from typing import Union

from .utils import base64url_decode


class Token:
    """
    The parsed PASETO token class.
    """

    def __init__(self, version: str, purpose: str, payload: bytes, footer: bytes = b""):
        self._version = version
        self._purpose = purpose
        self._payload = payload
        self._footer = footer
        self._header = (version + "." + purpose + ".").encode("utf-8")

    @classmethod
    def new(cls, token: Union[bytes, str]):
        token = token if isinstance(token, str) else token.decode("utf-8")
        t = token.split(".")
        if len(t) != 3 and len(t) != 4:
            raise ValueError("token is invalid.")
        p = base64url_decode(t[2])
        f = base64url_decode(t[3]) if len(t) == 4 else b""
        return cls(t[0], t[1], p, f)

    @property
    def version(self) -> str:
        return self._version

    @property
    def purpose(self) -> str:
        return self._purpose

    @property
    def header(self) -> bytes:
        return self._header

    @property
    def payload(self) -> bytes:
        return self._payload

    @payload.setter
    def payload(self, payload: bytes):
        self._payload = payload
        return

    @property
    def footer(self) -> bytes:
        return self._footer

    @footer.setter
    def footer(self, footer: bytes):
        self._footer = footer
        return



import hashlib
from secrets import token_bytes
from typing import Union

from Cryptodome.Cipher import ChaCha20
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)

from ..exceptions import DecryptError, EncryptError, SignError, VerifyError
from ..key_interface import KeyInterface
from ..utils import base64url_encode, pae


class V4Local(KeyInterface):
    def __init__(self, key: Union[str, bytes]):
        super().__init__("v4", "local", key)
        return

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:
        if nonce:
            if len(nonce) != 32:
                raise ValueError("nonce should be 32 bytes.")
        else:
            nonce = token_bytes(32)
        tmp = self._generate_hash(self._key, b"paseto-encryption-key" + nonce, 56)
        ek = tmp[0:32]
        n2 = tmp[32:]
        ak = self._generate_hash(self._key, b"paseto-auth-key-for-aead" + nonce, 32)

        try:
            cipher = ChaCha20.new(key=ek, nonce=n2)
            c = cipher.encrypt(payload)
            pre_auth = pae([self.header, nonce, c, footer, implicit_assertion])
            t = self._generate_hash(ak, pre_auth, 32)
            token = self._header + base64url_encode(nonce + c + t)
            if footer:
                token += b"." + base64url_encode(footer)
            return token
        except Exception as err:
            raise EncryptError("Failed to encrypt a message.") from err

    def decrypt(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        n = payload[0:32]
        c = payload[32 : len(payload) - 32]
        t = payload[-32:]
        tmp = self._generate_hash(self._key, b"paseto-encryption-key" + n, 56)
        ek = tmp[0:32]
        n2 = tmp[32:]
        ak = self._generate_hash(self._key, b"paseto-auth-key-for-aead" + n, 32)
        pre_auth = pae([self.header, n, c, footer, implicit_assertion])
        t2 = self._generate_hash(ak, pre_auth, 32)
        if t != t2:
            raise DecryptError("Hash value mismatch.")
        try:
            cipher = ChaCha20.new(key=ek, nonce=n2)
            return cipher.decrypt(c)
        except Exception as err:
            raise DecryptError("Failed to decrypt a message.") from err

    def _generate_hash(self, key: bytes, msg: bytes, size: int) -> bytes:
        try:
            h = hashlib.blake2b(key=key, digest_size=size)
            h.update(msg)
            return h.digest()
        except Exception as err:
            raise EncryptError("Failed to generate hash.") from err


class V4Public(KeyInterface):
    def __init__(self, key: Union[str, bytes]):
        super().__init__("v4", "public", key)
        self._sig_size = 64

        if not isinstance(self._key, (Ed25519PublicKey, Ed25519PrivateKey)):
            raise ValueError("The key is not Ed25519 key.")
        return

    def sign(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        if isinstance(self._key, Ed25519PublicKey):
            raise ValueError("A public key cannot be used for signing.")
        m2 = pae([self.header, payload, footer, implicit_assertion])
        try:
            return self._key.sign(m2)
        except Exception as err:
            raise SignError("Failed to sign.") from err

    def verify(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ):
        if len(payload) <= self._sig_size:
            raise ValueError("Invalid payload.")

        sig = payload[-self._sig_size :]
        m = payload[: len(payload) - self._sig_size]
        k = (
            self._key
            if isinstance(self._key, Ed25519PublicKey)
            else self._key.public_key()
        )
        m2 = pae([self.header, m, footer, implicit_assertion])
        try:
            k.verify(sig, m2)
        except Exception as err:
            raise VerifyError("Failed to verify.") from err
        return m



from .exceptions import DecryptError, EncryptError, PysetoError, SignError, VerifyError
from .key import Key
from .pyseto import decode, encode

__version__ = "0.2.0"
__title__ = "PySETO"
__description__ = "A Python implementation of PASETO"
__url__ = "https://pyseto.readthedocs.io"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"
__author__ = "AJITOMI Daisuke"
__email__ = "ajitomi@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2021 AJITOMI Daisuke"
__all__ = [
    "encode",
    "decode",
    "Key",
    "PysetoError",
    "SignError",
    "VerifyError",
    "EncryptError",
    "DecryptError",
]



from typing import List, Union

from .key_interface import KeyInterface
from .token import Token
from .utils import base64url_encode


def encode(
    key: KeyInterface,
    payload: Union[bytes, str],
    footer: Union[bytes, str] = b"",
    implicit_assertion: Union[bytes, str] = b"",
    nonce: bytes = b"",
) -> bytes:
    """
    Encodes a message to a PASETO token with a key.

    Args:
        key (KeyInterface): A key for encryption or signing.
        payload (Union[bytes, str]): A message to be encrypted or signed.
        footer (Union[bytes, str]): A footer.
        implicit_assertion (Union[bytes, str]): An implicit assertion.
        nonce (bytes): A nonce.
    Returns:
        str: A PASETO token.
    Raise:
        ValueError: Invalid arguments.
        EncryptError: Failed to encrypt the message.
        SignError: Failed to sign the message.
    """
    bp = payload if isinstance(payload, bytes) else payload.encode("utf-8")
    bf = footer if isinstance(footer, bytes) else footer.encode("utf-8")
    bi = (
        implicit_assertion
        if isinstance(implicit_assertion, bytes)
        else implicit_assertion.encode("utf-8")
    )
    if key.type == "local":
        return key.encrypt(bp, bf, bi, nonce)

    sig = key.sign(bp, bf, bi)
    token = key.header + base64url_encode(bp + sig)
    if bf:
        token += b"." + base64url_encode(bf)
    return token


def decode(
    keys: Union[KeyInterface, List[KeyInterface]],
    token: Union[bytes, str],
    implicit_assertion: Union[bytes, str] = b"",
) -> Token:
    """
    Decodes a PASETO token with a key.

    Args:
        key (KeyInterface): A key for decryption or verifying the signature in the token.
        token (Union[bytes, str]): A PASETO token to be decrypted or verified.
        implicit_assertion (Optional[Union[bytes, str]]): An implicit assertion.
    Returns:
        Token: A parsed PASETO token object.
    Raise:
        ValueError: Invalid arguments.
        DecryptError: Failed to decrypt the message.
        VerifyError: Failed to verify the message.
    """
    if isinstance(keys, list):
        raise ValueError("List[KeyInterface] for keys is not supported so far.")
    else:
        keys = [keys]
    bi = (
        implicit_assertion
        if isinstance(implicit_assertion, bytes)
        else implicit_assertion.encode("utf-8")
    )

    t = Token.new(token)
    for k in keys:
        if k.header != t.header:
            continue
        if k.type == "local":
            t.payload = k.decrypt(t.payload, t.footer, bi)
            return t
        t.payload = k.verify(t.payload, t.footer, bi)
        return t
    raise ValueError("key is not found for verifying the token.")



import hashlib
import hmac
from secrets import token_bytes
from typing import Union

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKey,
    EllipticCurvePublicKey,
)
from cryptography.hazmat.primitives.asymmetric.utils import (
    decode_dss_signature,
    encode_dss_signature,
)
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from ..exceptions import DecryptError, EncryptError, SignError, VerifyError
from ..key_interface import KeyInterface
from ..utils import base64url_encode, i2osp, os2ip, pae


class V3Local(KeyInterface):
    def __init__(self, key: Union[str, bytes]):
        super().__init__("v3", "local", key)
        return

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:
        if nonce:
            if len(nonce) != 32:
                raise ValueError("nonce should be 32 bytes.")
        else:
            nonce = token_bytes(32)
        e = HKDF(
            algorithm=hashes.SHA384(),
            length=48,
            salt=None,
            info=b"paseto-encryption-key" + nonce,
        )
        a = HKDF(
            algorithm=hashes.SHA384(),
            length=48,
            salt=None,
            info=b"paseto-auth-key-for-aead" + nonce,
        )
        try:
            tmp = e.derive(self._key)
            ek = tmp[0:32]
            n2 = tmp[32:]
            ak = a.derive(self._key)
        except Exception as err:
            raise DecryptError("Failed to derive keys.") from err

        try:
            c = Cipher(algorithms.AES(ek), modes.CTR(n2)).encryptor().update(payload)
            pre_auth = pae([self.header, nonce, c, footer, implicit_assertion])
            t = hmac.new(ak, pre_auth, hashlib.sha384).digest()
            token = self._header + base64url_encode(nonce + c + t)
            if footer:
                token += b"." + base64url_encode(footer)
            return token
        except Exception as err:
            raise EncryptError("Failed to encrypt a message.") from err

    def decrypt(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        n = payload[0:32]
        c = payload[32 : len(payload) - 48]
        t = payload[-48:]
        e = HKDF(
            algorithm=hashes.SHA384(),
            length=48,
            salt=None,
            info=b"paseto-encryption-key" + n,
        )
        a = HKDF(
            algorithm=hashes.SHA384(),
            length=48,
            salt=None,
            info=b"paseto-auth-key-for-aead" + n,
        )
        try:
            tmp = e.derive(self._key)
            ek = tmp[0:32]
            n2 = tmp[32:]
            ak = a.derive(self._key)
        except Exception as err:
            raise DecryptError("Failed to derive keys.") from err

        pre_auth = pae([self.header, n, c, footer, implicit_assertion])
        t2 = hmac.new(ak, pre_auth, hashlib.sha384).digest()
        if t != t2:
            raise DecryptError("Hash value mismatch.")

        try:
            return Cipher(algorithms.AES(ek), modes.CTR(n2)).decryptor().update(c)
        except Exception as err:
            raise DecryptError("Failed to decrypt a message.") from err


class V3Public(KeyInterface):
    def __init__(self, key: Union[str, bytes]):
        super().__init__("v3", "public", key)
        self._sig_size = 96

        if not isinstance(self._key, (EllipticCurvePublicKey, EllipticCurvePrivateKey)):
            raise ValueError("The key is not ECDSA key.")
        return

    def sign(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        if isinstance(self._key, EllipticCurvePublicKey):
            raise ValueError("A public key cannot be used for signing.")
        pk = self._public_key_compress(
            self._key.private_numbers().public_numbers.x,
            self._key.private_numbers().public_numbers.y,
        )
        m2 = pae([pk, self.header, payload, footer, implicit_assertion])
        try:
            sig = self._key.sign(m2, ec.ECDSA(hashes.SHA384()))
            return self._der_to_os(self._key.curve.key_size, sig)
        except Exception as err:
            raise SignError("Failed to sign.") from err

    def verify(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ):
        if len(payload) <= self._sig_size:
            raise ValueError("Invalid payload.")

        sig = payload[-self._sig_size :]
        m = payload[: len(payload) - self._sig_size]
        k = (
            self._key
            if isinstance(self._key, EllipticCurvePublicKey)
            else self._key.public_key()
        )
        pk = self._public_key_compress(k.public_numbers().x, k.public_numbers().y)
        m2 = pae([pk, self.header, m, footer, implicit_assertion])
        try:
            der_sig = self._os_to_der(self._key.curve.key_size, sig)
            k.verify(der_sig, m2, ec.ECDSA(hashes.SHA384()))
        except Exception as err:
            raise VerifyError("Failed to verify.") from err
        return m

    def _public_key_compress(self, x: int, y: int) -> bytes:
        bx = x.to_bytes((x.bit_length() + 7) // 8, byteorder="big")
        by = y.to_bytes((y.bit_length() + 7) // 8, byteorder="big")
        s = bytearray(1)
        s[0] = 0x02 + (by[len(by) - 1] & 1)
        # s: bytearray = [0x02 + (by[len(by) - 1] & 1)]
        return bytes(s) + bx

    def _der_to_os(self, key_size: int, sig: bytes) -> bytes:
        num_bytes = (key_size + 7) // 8
        r, s = decode_dss_signature(sig)
        return i2osp(r, num_bytes) + i2osp(s, num_bytes)

    def _os_to_der(self, key_size: int, sig: bytes) -> bytes:
        num_bytes = (key_size + 7) // 8
        if len(sig) != 2 * num_bytes:
            raise ValueError("Invalid signature.")
        r = os2ip(sig[:num_bytes])
        s = os2ip(sig[num_bytes:])
        return encode_dss_signature(r, s)



import hashlib
import hmac
from secrets import token_bytes
from typing import Union

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from ..exceptions import DecryptError, EncryptError, SignError, VerifyError
from ..key_interface import KeyInterface
from ..utils import base64url_encode, pae


class V1Local(KeyInterface):
    def __init__(self, key: Union[str, bytes]):
        super().__init__("v1", "local", key)
        return

    def encrypt(
        self,
        payload: bytes,
        footer: bytes = b"",
        implicit_assertion: bytes = b"",
        nonce: bytes = b"",
    ) -> bytes:
        n = self._get_nonce(payload, nonce)
        e = HKDF(
            algorithm=hashes.SHA384(),
            length=32,
            salt=n[0:16],
            info=b"paseto-encryption-key",
        )
        a = HKDF(
            algorithm=hashes.SHA384(),
            length=32,
            salt=n[0:16],
            info=b"paseto-auth-key-for-aead",
        )
        try:
            ek = e.derive(self._key)
            ak = a.derive(self._key)
        except Exception as err:
            raise DecryptError("Failed to derive keys.") from err

        try:
            c = (
                Cipher(algorithms.AES(ek), modes.CTR(n[16:]))
                .encryptor()
                .update(payload)
            )
            pre_auth = pae([self.header, n, c, footer])
            t = hmac.new(ak, pre_auth, hashlib.sha384).digest()
            token = self._header + base64url_encode(n + c + t)
            if footer:
                token += b"." + base64url_encode(footer)
            return token
        except Exception as err:
            raise EncryptError("Failed to encrypt a message.") from err

    def decrypt(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        n = payload[0:32]
        t = payload[-48:]
        c = payload[32 : len(payload) - 48]
        e = HKDF(
            algorithm=hashes.SHA384(),
            length=32,
            salt=n[0:16],
            info=b"paseto-encryption-key",
        )
        a = HKDF(
            algorithm=hashes.SHA384(),
            length=32,
            salt=n[0:16],
            info=b"paseto-auth-key-for-aead",
        )
        try:
            ek = e.derive(self._key)
            ak = a.derive(self._key)
        except Exception as err:
            raise DecryptError("Failed to derive keys.") from err

        pre_auth = pae([self.header, n, c, footer])
        t2 = hmac.new(ak, pre_auth, hashlib.sha384).digest()
        if t != t2:
            raise DecryptError("Hash value mismatch.")

        try:
            return Cipher(algorithms.AES(ek), modes.CTR(n[16:])).decryptor().update(c)
        except Exception as err:
            raise DecryptError("Failed to decrypt a message.") from err

    def _get_nonce(self, msg: bytes, nonce: bytes = b"") -> bytes:
        if nonce:
            if len(nonce) != 32:
                raise ValueError("nonce should be 32 bytes.")
        else:
            nonce = token_bytes(32)

        try:
            return hmac.new(nonce, msg, hashlib.sha384).digest()[0:32]
        except Exception as err:
            raise EncryptError("Failed to get nonce.") from err


class V1Public(KeyInterface):
    def __init__(self, key: Union[str, bytes]):
        super().__init__("v1", "public", key)
        self._sig_size = 256

        if not isinstance(self._key, (RSAPublicKey, RSAPrivateKey)):
            raise ValueError("The key is not RSA key.")

        self._padding = padding.PSS(mgf=padding.MGF1(hashes.SHA384()), salt_length=48)
        return

    def sign(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ) -> bytes:
        if isinstance(self._key, RSAPublicKey):
            raise ValueError("Public key cannot be used for signing.")
        m2 = pae([self.header, payload, footer])
        try:
            return self._key.sign(m2, self._padding, hashes.SHA384())
        except Exception as err:
            raise SignError("Failed to sign.") from err

    def verify(
        self, payload: bytes, footer: bytes = b"", implicit_assertion: bytes = b""
    ):
        if len(payload) <= self._sig_size:
            raise ValueError("Invalid payload.")

        sig = payload[-self._sig_size :]
        m = payload[: len(payload) - self._sig_size]
        k = self._key if isinstance(self._key, RSAPublicKey) else self._key.public_key()
        m2 = pae([self.header, m, footer])
        try:
            k.verify(sig, m2, self._padding, hashes.SHA384())
        except Exception as err:
            raise VerifyError("Failed to verify.") from err
        return m



class PysetoError(Exception):
    """
    Base class for all exceptions.
    """

    pass


class NotSupportedError(PysetoError):
    """
    An Exception occurred when the function is not supported for the key object.
    """

    pass


class EncryptError(PysetoError):
    """
    An Exception occurred when an encryption process failed.
    """

    pass


class DecryptError(PysetoError):
    """
    An Exception occurred when an decryption process failed.
    """

    pass


class SignError(PysetoError):
    """
    An Exception occurred when a signing process failed.
    """

    pass


class VerifyError(PysetoError):
    """
    An Exception occurred when a verification process failed.
    """

    pass



from typing import Union

from .versions.v1 import V1Local, V1Public
from .versions.v2 import V2Local, V2Public
from .versions.v3 import V3Local, V3Public
from .versions.v4 import V4Local, V4Public


class Key:
    @staticmethod
    def new(version: str, type: str, key: Union[bytes, str] = b""):
        if type != "local" and type != "public":
            raise ValueError(f"Invalid type: {type}")
        if version == "v1":
            return V1Public(key) if type == "public" else V1Local(key)
        if version == "v2":
            return V2Public(key) if type == "public" else V2Local(key)
        if version == "v3":
            return V3Public(key) if type == "public" else V3Local(key)
        if version == "v4":
            return V4Public(key) if type == "public" else V4Local(key)
        raise ValueError(f"Invalid version: {version}")
